@@ -2,36 +2,16 @@
 
 from __future__ import absolute_import, unicode_literals
 
-# Import the fastest implementation of
-# pickle package. This should be removed
-# when python3 come the unique supported
-# python version
-try:
-    import cPickle as pickle
-except ImportError:
-    import pickle
-
 import random
 import socket
 import warnings
 import zlib
-
-try:
-    from django.utils.encoding import smart_bytes
-    from django.utils.encoding import smart_text
-except ImportError:
-    from django.utils.encoding import smart_str as smart_bytes
-    from django.utils.encoding import smart_unicode as smart_text
+from collections import OrderedDict
 
 from django.conf import settings
-from django.core.cache.backends.base import get_key_func
+from django.core.cache.backends.base import DEFAULT_TIMEOUT, get_key_func
 from django.core.exceptions import ImproperlyConfigured
-from django.utils.datastructures import SortedDict
-
-try:
-    from django.core.cache.backends.base import DEFAULT_TIMEOUT
-except ImportError:
-    DEFAULT_TIMEOUT = object()
+from django.utils.encoding import smart_text
 
 from redis.exceptions import ConnectionError
 from redis.exceptions import ResponseError
@@ -39,19 +19,18 @@ from redis.exceptions import ResponseError
 # Compatibility with redis-py 2.10.x+
 
 try:
-    from redis.exceptions import TimeoutError
-    _main_exceptions = (TimeoutError, ConnectionError, socket.timeout)
+    from redis.exceptions import TimeoutError, ResponseError
+    _main_exceptions = (TimeoutError, ResponseError, ConnectionError, socket.timeout)
 except ImportError:
     _main_exceptions = (ConnectionError, socket.timeout)
 
-from ..util import CacheKey, integer_types
-from ..exceptions import ConnectionInterrupted
+from ..util import CacheKey, load_class, integer_types
+from ..exceptions import ConnectionInterrupted, CompressorError
 from .. import pool
 
 
 class DefaultClient(object):
     def __init__(self, server, params, backend):
-        self._pickle_version = -1
         self._backend = backend
         self._server = server
         self._params = params
@@ -67,11 +46,16 @@ class DefaultClient(object):
 
         self._clients = [None] * len(self._server)
         self._options = params.get("OPTIONS", {})
-        self._options.setdefault("COMPRESS_COMPRESSOR", zlib.compress)
-        self._options.setdefault("COMPRESS_DECOMPRESSOR", zlib.decompress)
-        self._options.setdefault("COMPRESS_DECOMPRESSOR_ERROR", zlib.error)
 
-        self.setup_pickle_version()
+        serializer_path = self._options.get("SERIALIZER", "django_redis.serializers.pickle.PickleSerializer")
+        serializer_cls = load_class(serializer_path)
+
+        compressor_path = self._options.get("COMPRESSOR", "django_redis.compressors.identity.IdentityCompressor")
+        compressor_cls = load_class(compressor_path)
+
+        self._serializer = serializer_cls(options=self._options)
+        self._compressor = compressor_cls(options=self._options);
+
         self.connection_factory = pool.get_connection_factory(options=self._options)
 
     def __contains__(self, key):
@@ -114,13 +98,6 @@ class DefaultClient(object):
         """
         return self.connection_factory.connect(self._server[index])
 
-    def setup_pickle_version(self):
-        if "PICKLE_VERSION" in self._options:
-            try:
-                self._pickle_version = int(self._options["PICKLE_VERSION"])
-            except (ValueError, TypeError):
-                raise ImproperlyConfigured("PICKLE_VERSION value must be an integer")
-
     def set(self, key, value, timeout=DEFAULT_TIMEOUT, version=None, client=None, nx=False, xx=False):
         """
         Persist a value to the cache, and set an optional expiration time.
@@ -131,7 +108,7 @@ class DefaultClient(object):
             client = self.get_client(write=True)
 
         nkey = self.make_key(key, version=version)
-        nvalue = self.pickle(value)
+        nvalue = self.encode(value)
 
         if timeout is True:
             warnings.warn("Using True as timeout value, is now deprecated.", DeprecationWarning)
@@ -143,10 +120,9 @@ class DefaultClient(object):
         try:
             if timeout is not None:
                 if timeout > 0:
-                    timeout = int(timeout)
-                elif timeout == 0:
-                    timeout = None
-                elif timeout < 0:
+                    # Convert to milliseconds
+                    timeout = int(timeout * 1000)
+                elif timeout <= 0:
                     if nx:
                         # Using negative timeouts when nx is True should
                         # not expire (in our case delete) the value if it exists.
@@ -158,7 +134,7 @@ class DefaultClient(object):
                         # than to set it and than expire in a pipeline
                         return self.delete(key, client=client, version=version)
 
-            return client.set(nkey, nvalue, nx=nx, ex=timeout, xx=xx)
+            return client.set(nkey, nvalue, nx=nx, px=timeout, xx=xx)
         except _main_exceptions as e:
             raise ConnectionInterrupted(connection=client, parent=e)
 
@@ -206,7 +182,7 @@ class DefaultClient(object):
         """
         Retrieve a value from the cache.
 
-        Returns unpickled value if key is found, the default if not.
+        Returns decoded value if key is found, the default if not.
         """
         if client is None:
             client = self.get_client(write=False)
@@ -221,9 +197,36 @@ class DefaultClient(object):
         if value is None:
             return default
 
-        return self.unpickle(value)
+        return self.decode(value)
 
-    def delete(self, key, version=None, client=None):
+    def persist(self, key, version=None, client=None):
+        if client is None:
+            client = self.get_client(write=True)
+
+        key = self.make_key(key, version=version)
+
+        if client.exists(key):
+            client.persist(key)
+
+    def expire(self, key, timeout, version=None, client=None):
+        if client is None:
+            client = self.get_client(write=True)
+
+        key = self.make_key(key, version=version)
+
+        if client.exists(key):
+            client.expire(key, timeout)
+
+    def lock(self, key, version=None, timeout=None, sleep=0.1,
+             blocking_timeout=None, client=None):
+        if client is None:
+            client = self.get_client(write=True)
+
+        key = self.make_key(key, version=version)
+        return client.lock(key, timeout=timeout, sleep=sleep,
+                           blocking_timeout=blocking_timeout)
+
+    def delete(self, key, version=None, prefix=None, client=None):
         """
         Remove a key from the cache.
         """
@@ -231,11 +234,12 @@ class DefaultClient(object):
             client = self.get_client(write=True)
 
         try:
-            return client.delete(self.make_key(key, version=version))
+            return client.delete(self.make_key(key, version=version,
+                                               prefix=prefix))
         except _main_exceptions as e:
             raise ConnectionInterrupted(connection=client, parent=e)
 
-    def delete_pattern(self, pattern, version=None, client=None):
+    def delete_pattern(self, pattern, version=None, prefix=None, client=None):
         """
         Remove all keys matching pattern.
         """
@@ -243,12 +247,13 @@ class DefaultClient(object):
         if client is None:
             client = self.get_client(write=True)
 
-        pattern = self.make_key(pattern, version=version)
+        pattern = self.make_key(pattern, version=version, prefix=prefix)
         try:
-            keys = client.keys(pattern)
-
-            if keys:
-                return client.delete(*keys)
+            count = 0
+            for key in client.scan_iter(pattern):
+                client.delete(key)
+                count += 1
+            return count
         except _main_exceptions as e:
             raise ConnectionInterrupted(connection=client, parent=e)
 
@@ -260,10 +265,11 @@ class DefaultClient(object):
         if client is None:
             client = self.get_client(write=True)
 
+        keys = [self.make_key(k, version=version) for k in keys]
+
         if not keys:
             return
 
-        keys = [self.make_key(k, version=version) for k in keys]
         try:
             return client.delete(*keys)
         except _main_exceptions as e:
@@ -273,42 +279,43 @@ class DefaultClient(object):
         """
         Flush all cache keys.
         """
+
         if client is None:
             client = self.get_client(write=True)
 
-        client.flushdb()
+        try:
+            count = 0
+            for key in client.scan_iter("*"):
+                client.delete(key)
+                count += 1
+            return count
+        except _main_exceptions as e:
+            raise ConnectionInterrupted(connection=client, parent=e)
 
-    def unpickle(self, value):
+    def decode(self, value):
         """
-        Unpickles the given value.
+        Decode the given value.
         """
         try:
             value = int(value)
         except (ValueError, TypeError):
-            if self._options.get("COMPRESS_MIN_LEN", 0) > 0:
-                try:
-                    value = self._options["COMPRESS_DECOMPRESSOR"](value)
-                except self._options["COMPRESS_DECOMPRESSOR_ERROR"]:
-                    # Handle little values, chosen to be not compressed
-                    pass
-            value = pickle.loads(smart_bytes(value))
+            try:
+                value = self._compressor.decompress(value)
+            except CompressorError:
+                # Handle little values, chosen to be not compressed
+                pass
+            value = self._serializer.loads(value)
         return value
 
-    def pickle(self, value):
+    def encode(self, value):
         """
-        Pickle the given value.
+        Encode the given value.
         """
 
         if isinstance(value, bool) or not isinstance(value, integer_types):
-            pickled_value = pickle.dumps(value, self._pickle_version)
-            if self._options.get("COMPRESS_MIN_LEN", 0) > 0:
-                if len(pickled_value) >= self._options["COMPRESS_MIN_LEN"]:
-                    # We should try to compress if COMPRESS_MIN_LEN > 0
-                    # and this string is longer than our min threshold.
-                    compressed = self._options["COMPRESS_COMPRESSOR"](pickled_value)
-                    if len(compressed) < len(pickled_value):
-                        pickled_value = compressed
-            return pickled_value
+            value = self._serializer.dumps(value)
+            value = self._compressor.compress(value)
+            return value
 
         return value
 
@@ -323,7 +330,7 @@ class DefaultClient(object):
         if not keys:
             return {}
 
-        recovered_data = SortedDict()
+        recovered_data = OrderedDict()
 
         new_keys = [self.make_key(k, version=version) for k in keys]
         map_keys = dict(zip(new_keys, keys))
@@ -336,7 +343,7 @@ class DefaultClient(object):
         for key, value in zip(new_keys, results):
             if value is None:
                 continue
-            recovered_data[map_keys[key]] = self.unpickle(value)
+            recovered_data[map_keys[key]] = self.decode(value)
         return recovered_data
 
     def set_many(self, data, timeout=DEFAULT_TIMEOUT, version=None, client=None):
@@ -365,19 +372,32 @@ class DefaultClient(object):
         key = self.make_key(key, version=version)
 
         try:
-            if not client.exists(key):
-                raise ValueError("Key '%s' not found" % key)
-
             try:
-                value = client.incr(key, delta)
+                # if key expired after exists check, then we get
+                # key with wrong value and ttl -1.
+                # use lua script for atomicity
+                lua = """
+                local exists = redis.call('EXISTS', KEYS[1])
+                if (exists == 1) then
+                    return redis.call('INCRBY', KEYS[1], ARGV[1])
+                else return false end
+                """
+                value = client.eval(lua, 1, key, delta)
+                if value is None:
+                    raise ValueError("Key '%s' not found" % key)
             except ResponseError:
                 # if cached value or total value is greater than 64 bit signed
                 # integer.
-                # elif int is pickled. so redis sees the data as string.
+                # elif int is encoded. so redis sees the data as string.
                 # In this situations redis will throw ResponseError
 
                 # try to keep TTL of key
+
                 timeout = client.ttl(key)
+                # returns -2 if the key does not exist
+                # means, that key have expired
+                if timeout == -2:
+                    raise ValueError("Key '%s' not found" % key)
                 value = self.get(key, version=version, client=client) + delta
                 self.set(key, value, version=version, timeout=timeout,
                          client=client)
@@ -414,7 +434,16 @@ class DefaultClient(object):
             return 0
 
         t = client.ttl(key)
-        return (t >= 0 and t or None)
+
+        if t >= 0:
+            return t
+        elif t == -1:
+            return None
+        elif t == -2:
+            return 0
+        else:
+            # Should never reach here
+            return None
 
     def has_key(self, key, version=None, client=None):
         """
@@ -462,10 +491,17 @@ class DefaultClient(object):
         except _main_exceptions as e:
             raise ConnectionInterrupted(connection=client, parent=e)
 
-    def make_key(self, key, version=None):
+    def make_key(self, key, version=None, prefix=None):
         if isinstance(key, CacheKey):
             return key
-        return CacheKey(self._backend.make_key(key, version))
+
+        if prefix is None:
+            prefix = self._backend.key_prefix
+
+        if version is None:
+            version = self._backend.version
+
+        return CacheKey(self._backend.key_func(key, prefix, version))
 
     def close(self, **kwargs):
         if getattr(settings, "DJANGO_REDIS_CLOSE_CONNECTION", False):
