@@ -3,16 +3,12 @@
 from __future__ import absolute_import, unicode_literals
 
 import re
+from collections import OrderedDict
 
 from redis.exceptions import ConnectionError
 
 from django.conf import settings
-from django.utils.datastructures import SortedDict
-
-try:
-    from django.utils.encoding import smart_text
-except ImportError:
-    from django.utils.encoding import smart_unicode as smart_text
+from django.utils.encoding import smart_text
 
 from ..hash_ring import HashRing
 from ..exceptions import ConnectionInterrupted
@@ -73,7 +69,7 @@ class ShardClient(DefaultClient):
         if not keys:
             return {}
 
-        recovered_data = SortedDict()
+        recovered_data = OrderedDict()
 
         new_keys = [self.make_key(key, version=version) for key in keys]
         map_keys = dict(zip(new_keys, keys))
@@ -133,6 +129,44 @@ class ShardClient(DefaultClient):
 
         return super(ShardClient, self).delete(key=key, version=version, client=client)
 
+    def ttl(self, key, version=None, client=None):
+        """
+        Executes TTL redis command and return the "time-to-live" of specified key.
+        If key is a non volatile key, it returns None.
+        """
+
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super(ShardClient, self).ttl(key=key, version=version, client=client)
+
+    def persist(self, key, version=None, client=None):
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super(ShardClient, self).persist(key=key, version=version, client=client)
+
+    def expire(self, key, timeout, version=None, client=None):
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super(ShardClient, self).expire(key=key, timeout=timeout,
+                                               version=version, client=client)
+
+    def lock(self, key, version=None, timeout=None, sleep=0.1,
+             blocking_timeout=None, client=None):
+
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        key = self.make_key(key, version=version)
+        return super(ShardClient, self).lock(key, timeout=timeout, sleep=sleep, client=client,
+                                             blocking_timeout=blocking_timeout)
+
     def delete_many(self, keys, version=None):
         """
         Remove multiple keys at once.
@@ -204,7 +238,7 @@ class ShardClient(DefaultClient):
         decoded_keys = (smart_text(k) for k in keys)
         return [self.reverse_key(k) for k in decoded_keys]
 
-    def delete_pattern(self, pattern, version=None):
+    def delete_pattern(self, pattern, version=None, client=None):
         """
         Remove all keys matching pattern.
         """
@@ -213,7 +247,7 @@ class ShardClient(DefaultClient):
 
         keys = []
         for server, connection in self._serverdict.items():
-            keys.extend(connection.keys(pattern))
+            keys.extend([key for key in connection.scan_iter(pattern)])
 
         res = 0
         if keys:
